import functools
from typing import Tuple

import numba
import numpy as np
import scipy.interpolate
import scipy.sparse
import scipy.special
from numpy.typing import NDArray


@functools.lru_cache
def xw_lgl(num_point: int) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Compute the Legendre-Gauss-Lobatto nodes and weights.
    c.f. https://math.stackexchange.com/questions/2058111/how-to-calculate-nodes-and-weights-of-legendre-gauss-lobatto-rule
    """
    if num_point == 1:
        return np.array([0], dtype=np.float64), np.array([2], dtype=np.float64)
    n = num_point - 1
    poly = scipy.special.legendre(n)
    deriv = np.polyder(poly)
    x = [-1.]
    roots = np.roots(deriv)
    for root in roots:
        x.append(root.real)
    x.append(1.)
    x.sort()
    w = [2. / n / (n + 1)]
    for x_i in x[1:-1]:
        w.append(2. / n / (n + 1) / np.polyval(poly, x_i) ** 2)
    w.append(2. / n / (n + 1))
    return np.array(x, dtype=np.float64), np.array(w, dtype=np.float64)


@functools.lru_cache
def D_lgl(num_point: int) -> NDArray[np.float64]:
    """Compute the derivative matrix of the Legendre-Gauss-Lobatto nodes, i.e. x' = D @ x"""
    x, _ = xw_lgl(num_point)
    D = []
    for i in range(num_point):
        y = np.zeros(num_point)
        y[i] = 1
        poly = scipy.interpolate.lagrange(x, y)
        deriv_poly = np.polyder(poly)
        D.append(np.polyval(deriv_poly, x))
    return np.array(D, dtype=np.float64).T


def lr_c(num_point: NDArray[np.int32]) -> Tuple[NDArray[np.int32], NDArray[np.int32]]:
    """Compute the left and right index of continuous variables"""
    l = np.concatenate([[0], np.cumsum(num_point[:-1] - 1)])
    return l, l + num_point


def lr_nc(num_point: NDArray[np.int32]) -> Tuple[NDArray[np.int32], NDArray[np.int32]]:
    """Compute the left and right index of non-continuous variables"""
    return np.concatenate(([0], np.cumsum(num_point[:-1]))), np.cumsum(num_point)


def xw_c(mesh: NDArray[np.float64], num_point: NDArray[np.int32]) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Compute the discretization nodes and weights of continuous variables"""
    l_c, r_c = lr_c(num_point)
    L_c = r_c[-1]  # length
    x = np.zeros(L_c)
    w = np.zeros(L_c)
    width = np.diff(mesh)  # length of each interval
    mid = (mesh[1:] + mesh[:-1]) / 2  # mid-point of each interval
    for l, r, n, d, m in zip(l_c, r_c, num_point, width, mid):
        x_, w_ = xw_lgl(n)
        x[l:r] = x_ * d / 2 + m
        w[l:r] += w_ * d / 2  # += since overlap nodes
    return x, w


def xw_nc(mesh: NDArray[np.float64], num_point: NDArray[np.int32]) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Compute the discretization nodes and weights of non-continuous variables"""
    l_c, r_c = lr_nc(num_point)
    L_c = r_c[-1]  # length
    x = np.zeros(L_c)
    w = np.zeros(L_c)
    width = np.diff(mesh)  # length of each interval
    mid = (mesh[1:] + mesh[:-1]) / 2  # mid-point of each interval
    for l, r, n, d, m in zip(l_c, r_c, num_point, width, mid):
        x_, w_ = xw_lgl(n)
        x[l:r] = x_ * d / 2 + m
        w[l:r] = w_ * d / 2
    return x, w


def c2nc(num_point: NDArray[np.int32]):
    """Return a closure that Convert continuous discretization to non-continuous discretization"""
    l_c, r_c = lr_c(num_point)
    l_nc, r_nc = lr_nc(num_point)
    L_nc = r_nc[-1]  # length

    @numba.njit
    def c2nc_(c: NDArray[np.float64]) -> NDArray[np.float64]:
        shape = c.shape
        shape = (L_nc,) + shape[1:]
        nc = np.zeros(shape)
        for l_c_, r_c_, l_nc_, r_nc_ in zip(l_c, r_c, l_nc, r_nc):
            nc[l_nc_:r_nc_] = c[l_c_:r_c_]
        return nc

    return c2nc_


def nc2c(num_point: NDArray[np.int32]):
    """Return a closure that Convert non-continuous discretization to continuous discretization"""
    l_c, r_c = lr_c(num_point)
    l_nc, r_nc = lr_nc(num_point)
    L_c = r_c[-1]  # length

    @numba.njit
    def nc2c_(nc: NDArray[np.float64]) -> NDArray[np.float64]:
        shape = nc.shape
        shape = (L_c,) + shape[1:]
        c = np.zeros(shape)
        for l_c_, r_c_, l_nc_, r_nc_ in zip(l_c, r_c, l_nc, r_nc):
            c[l_c_:r_c_] += nc[l_nc_:r_nc_]  # += since overlap nodes
        return c

    return nc2c_


def D_c(mesh: NDArray[np.float64], num_point: NDArray[np.int32]) -> NDArray[np.float64]:
    """Compute the derivative matrix of continuous variables"""
    data = []
    row = []
    col = []
    l_row, r_row = lr_nc(num_point)
    l_col, r_col = lr_c(num_point)
    L_row = r_row[-1]
    L_col = r_col[-1]
    width = np.diff(mesh)  # length of each interval
    for l_r, l_c, n, d in zip(l_row, l_col, num_point, width):
        D = D_lgl(n) / d * 2
        data.extend(D.flatten())
        row.extend(l_r + np.repeat(np.arange(n), n))
        col.extend(l_c + np.tile(np.arange(n), n))
    D_coo = scipy.sparse.coo_array((data, (row, col)), shape=(L_row, L_col))
    return D_coo.tocsr()


def D_nc(mesh: NDArray[np.float64], num_point: NDArray[np.int32]) -> NDArray[np.float64]:
    """Compute the derivative matrix of non-continuous variables"""
    data = []
    row = []
    col = []
    l_row, r_row = lr_nc(num_point)
    l_col, r_col = lr_nc(num_point)
    L_row = r_row[-1]
    L_col = r_col[-1]
    width = np.diff(mesh)  # length of each interval
    for l_r, l_c, n, d in zip(l_row, l_col, num_point, width):
        D = D_lgl(n) / d * 2
        data.extend(D.flatten())
        row.extend(l_r + np.repeat(np.arange(n), n))
        col.extend(l_c + np.tile(np.arange(n), n))
    D_coo = scipy.sparse.coo_array((data, (row, col)), shape=(L_row, L_col))
    return D_coo.tocsr()


def lr_v(num_point: NDArray[np.int32], continuity_xu: NDArray[np.bool]) \
        -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Return the left and right index of each variable before c2nc"""
    l_c, r_c = lr_c(num_point)
    l_nc, r_nc = lr_nc(num_point)
    L_c = r_c[-1]
    L_nc = r_nc[-1]
    l = [0]
    r = []
    for c in continuity_xu:
        if c:
            r.append(l[-1] + L_c)
            l.append(r[-1])
        else:
            r.append(l[-1] + L_nc)
            l.append(r[-1])
    return np.array(l[:-1]), np.array(r)


def lr_m(num_point: NDArray[np.int32], continuity_xu: NDArray[np.bool]) \
        -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Return the left and right index of each variable after c2nc (i.e. in middle stage)"""
    L_nc = lr_nc(num_point)[1][-1]
    n_xu = len(continuity_xu)
    return L_nc * np.arange(n_xu, dtype=np.int32), L_nc * (np.arange(n_xu, dtype=np.int32) + 1)


def v2m(num_point: NDArray[np.int32], continuity_xu: NDArray[np.bool]):
    """Return a closure that convert overlapped variables to non-overlap (middle stage) variables"""
    l_v, r_v = lr_v(num_point, continuity_xu)
    l_m, r_m = lr_m(num_point, continuity_xu)
    n_xu = len(continuity_xu)
    f_c2nc = c2nc(num_point)

    @numba.njit
    def v2m_(v: NDArray[np.float64]) -> NDArray[np.float64]:
        shape = v.shape
        shape = (r_m[-1],) + shape[1:]
        y = np.zeros(shape)
        for i in range(n_xu):
            if continuity_xu[i]:
                y[l_m[i]:r_m[i]] = f_c2nc(v[l_v[i]:r_v[i]])
            else:
                y[l_m[i]:r_m[i]] = v[l_v[i]:r_v[i]]
        return y

    return v2m_


def m2v(num_point: NDArray[np.int32], continuity_xu: NDArray[np.bool]):
    """Return a closure that convert middle stage variables to overlapped variables"""
    l_v, r_v = lr_v(num_point, continuity_xu)
    l_m, r_m = lr_m(num_point, continuity_xu)
    n_xu = len(continuity_xu)
    f_nc2c = nc2c(num_point)

    @numba.njit
    def m2v_(m: NDArray[np.float64]) -> NDArray[np.float64]:
        shape = m.shape
        shape = (r_v[-1],) + shape[1:]
        v = np.zeros(shape)
        for i in range(n_xu):
            if continuity_xu[i]:
                v[l_v[i]:r_v[i]] = f_nc2c(m[l_m[i]:r_m[i]])
            else:
                v[l_v[i]:r_v[i]] = m[l_m[i]:r_m[i]]
        return v

    return m2v_


def D_x(mesh: NDArray[np.float64], num_point: NDArray[np.int32], continuity_x: NDArray[np.bool]) \
        -> NDArray[np.float64]:
    """Compute the derivative matrix of all state variables"""
    D_c_ = D_c(mesh, num_point)
    D_nc_ = D_nc(mesh, num_point)
    diag = [D_c_ if c else D_nc_ for c in continuity_x]
    return scipy.sparse.block_diag(diag, format='csr')



from typing import Union, List, Iterable

import numpy as np
import sympy as sp
from numpy.typing import NDArray

from .fastfunc import FastFunc
from .phase import Phase


class System:
    def __init__(self, static_parameter: Union[int, List[str]], /, simplify: bool = False, parallel: bool = False,
                 fastmath: bool = False):
        """Initialize a system with given static parameters.
        If static_parameter is an integer, the static parameters will be named as s_0, ..., s_{n - 1}.

        If simplify is True, every symbolic expression will be simplified (by sympy.simplify) before being compiled.
        This may slow down the speed of compilation.

        If parallel is True, parallel flag will be passed to numba JIT compiler,
        which will generate parallel code for multicore CPUs.
        This may slow down the speed of compilation.

        If fastmath is True, fastmath flag will be passed to numba JIT compiler,
        see numba & LLVM documentations for details.

        Args:
            static_parameter (Union[int, List[str]]): Number of static parameters or list of static parameter names.
            simplify (bool, optional): Whether to use sympy to simplify Exprs. Defaults to False.
            parallel (bool, optional): Whether to use numba parallel mode. Defaults to False.
            fastmath (bool, optional): Whether to use numba fastmath mode. Defaults to False.
        """
        if isinstance(static_parameter, int):
            self._num_static_parameter = static_parameter
            self._name_static_parameter = [f's_{i}' for i in range(static_parameter)]
        elif isinstance(static_parameter, list):
            self._name_static_parameter = static_parameter
            self._num_static_parameter = len(static_parameter)
        else:
            raise ValueError('static_parameter must be int or list of str')

        self._symbol_static_parameter = [sp.Symbol(name) for name in self._name_static_parameter]
        self._symbols = self._symbol_static_parameter

        self._identifier_phase = 0

        self._simplify = simplify
        self._parallel = parallel
        self._fastmath = fastmath

        self._phase_set = False
        self._objective_set = False  # user must set an objective function
        self._system_constraint_set = False
        self.set_phase([])  # no phase is a valid system and is set by default
        self.set_system_constraint([], np.array([]), np.array([]))  # no system constraint by default

    def new_phase(self, state: Union[int, List[str]], control: Union[int, List[str]]):
        """Create a new phase for the given system.

        Args:
            state (Union[int, List[str]]): Number of state variables or list of state variable names.
            control (Union[int, List[str]]): Number of control variables or list of control variable names.

        Returns:
            Phase: A new phase, with static_parameter and configs correctly set.
        """
        self._identifier_phase += 1
        return Phase(state, control, self._symbol_static_parameter, self._identifier_phase - 1, simplify=self._simplify,
                     parallel=self._parallel, fastmath=self._fastmath)

    def set_phase(self, phase: List[Phase]):
        """Set the phases of the system.

        Args:
            phase (Phase): Phases of the system.
        """
        for i, p in enumerate(phase):
            if not p.ok:
                raise ValueError(f'Dynamics, boundary conditions, '
                                 f'or discretization scheme of phase {i} are not fully set')
        self._phase = phase
        self._num_phase = len(phase)

        self._phase_set = True

        self._update_lr_phase()
        self._update_symbols()
        if self._objective_set and self._system_constraint_set:
            self._update_jacobian_structure()
            self._update_hessian_structure()
        if self._system_constraint_set:
            self._update_bounds()
        return self

    def set_objective(self, objective: Union[float, sp.Expr]):
        """Set the objective of the system.

        Args:
            objective (Union[int, float, sp.Expr]): Objective of the system composed by s & I.
        """
        self._expr_objective = sp.sympify(objective)
        self._func_objective = FastFunc(self._expr_objective, self._symbols, self._simplify, self._parallel,
                                        self._fastmath)
        self._objective_set = True

        if self._phase_set and self._system_constraint_set:
            self._update_jacobian_structure()
            self._update_hessian_structure()
        return self

    def set_system_constraint(self, constraint: List[sp.Expr], lower_bound: Iterable[float],
                              upper_bound: Iterable[float]):
        """Set the system constraint of the system.

        Args:
            constraint (List[Union[int, float, sp.Expr]]): Global constraint of the system composed by s & I.
            lower_bound (NDArray[np.float64]): Lower bound of the constraints.
            upper_bound (NDArray[np.float64]): Upper bound of the constraints.
        """
        lower_bound = list(lower_bound)
        upper_bound = list(upper_bound)
        if not len(constraint) == len(lower_bound) == len(upper_bound):
            raise ValueError('constraint, lower_bound and upper_bound must have the same length')

        self._static_parameter_bounds_system = []
        self._expr_system_constraint = []
        lower_bound_system_constraint = []
        upper_bound_system_constraint = []
        for c, lb, ub in zip(constraint, lower_bound, upper_bound):
            if c.is_symbol and c in self.s:
                self._static_parameter_bounds_system.append((self.s.index(c), lb, ub))
            else:
                self._expr_system_constraint.append(sp.sympify(c))
                lower_bound_system_constraint.append(lb)
                upper_bound_system_constraint.append(ub)

        self._func_system_constraint = [FastFunc(c, self._symbols, self._simplify, self._parallel, self._fastmath)
                                        for c in self._expr_system_constraint]
        self._num_system_constraint = len(self._expr_system_constraint)
        self._lower_bound_system_constraint = np.array(lower_bound_system_constraint)
        self._upper_bound_system_constraint = np.array(upper_bound_system_constraint)

        self._system_constraint_set = True

        if self._phase_set:
            self._update_bounds()
        if self._objective_set and self._phase_set:
            self._update_jacobian_structure()
            self._update_hessian_structure()
        return self

    def update(self):
        """Update the system after changing discretization scheme of any phase.
        """
        if not self.ok:
            raise ValueError('System is not fully set')
        self._update_lr_phase()
        self._update_symbols()
        self._update_bounds()
        self._update_jacobian_structure()
        self._update_hessian_structure()

    def _update_lr_phase(self):
        if self.n_p == 0:
            self.l_p = np.array([], dtype=np.int32)
            self.r_p = np.array([], dtype=np.int32)
            self.l_s = 0
            self.r_s = self.n_s
        else:
            l_p = [0]
            r_p = []
            for p in self.p:
                r_p.append(l_p[-1] + p.L)
                l_p.append(r_p[-1])
            self.l_p = np.array(l_p[:-1], dtype=np.int32)
            self.r_p = np.array(r_p, dtype=np.int32)
            self.l_s = r_p[-1]
            self.r_s = self.l_s + self.n_s

    def _update_symbols(self):
        self._symbols = self._symbol_static_parameter
        mapping_integral = []
        for i, p in enumerate(self.p):
            self._symbols += p.I
            mapping_integral.extend([(i, j) for j in range(p.n_I)])
        self._num_symbol = len(self._symbols)
        self._mapping_integral = mapping_integral

    def _update_bounds(self):
        lower_bound_static_parameter = np.full(self.n_s, -np.inf, dtype=np.float64)
        upper_bound_static_parameter = np.full(self.n_s, np.inf, dtype=np.float64)
        for p in self.p:
            for i, lb, ub in p.s_b:
                lower_bound_static_parameter[i] = np.maximum(lower_bound_static_parameter[i], lb)
                upper_bound_static_parameter[i] = np.minimum(upper_bound_static_parameter[i], ub)
        for i, lb, ub in self._static_parameter_bounds_system:
            lower_bound_static_parameter[i] = np.maximum(lower_bound_static_parameter[i], lb)
            upper_bound_static_parameter[i] = np.minimum(upper_bound_static_parameter[i], ub)

        self._lower_bound_variable = \
            np.concatenate([p.v_lb for p in self.p] + [np.array(lower_bound_static_parameter)])
        self._upper_bound_variable = \
            np.concatenate([p.v_ub for p in self.p] + [np.array(upper_bound_static_parameter)])

        lower_bound_constraint = []
        upper_bound_constraint = []
        # system constraint
        lower_bound_constraint.append(self._lower_bound_system_constraint)
        upper_bound_constraint.append(self._upper_bound_system_constraint)
        # dynamics
        lower_bound_constraint.extend(np.zeros(p.n_x * p.l0_m, dtype=np.float64) for p in self.p)
        upper_bound_constraint.extend(np.zeros(p.n_x * p.l0_m, dtype=np.float64) for p in self.p)
        # phase constraint
        lower_bound_constraint.extend(np.repeat(p.c_lb, p.l0_m) for p in self.p)
        upper_bound_constraint.extend(np.repeat(p.c_ub, p.l0_m) for p in self.p)
        # boundary condition
        lower_bound_constraint.extend(np.zeros(p.n_bc, dtype=np.float64) for p in self.p)
        upper_bound_constraint.extend(np.zeros(p.n_bc, dtype=np.float64) for p in self.p)

        self._lower_bound_constraint = np.concatenate(lower_bound_constraint)
        self._upper_bound_constraint = np.concatenate(upper_bound_constraint)

    @property
    def n_s(self) -> int:
        """Number of static parameters."""
        return self._num_static_parameter

    @property
    def s(self) -> List[sp.Symbol]:
        """Sympy symbols of static parameters."""
        return self._symbol_static_parameter

    @property
    def n_p(self) -> int:
        """Number of phases."""
        return self._num_phase

    @property
    def p(self) -> List[Phase]:
        """Phases of system."""
        return self._phase

    @property
    def F_o(self) -> FastFunc:
        """FastFuncs of objective function."""
        return self._func_objective

    @property
    def n_c(self) -> int:
        """Number of system constraints."""
        return self._num_system_constraint

    @property
    def F_c(self) -> List[FastFunc]:
        """FastFuncs of system constraints."""
        return self._func_system_constraint

    @property
    def v_lb(self) -> NDArray[np.float64]:
        """Lower bound of variables."""
        return self._lower_bound_variable

    @property
    def v_ub(self) -> NDArray[np.float64]:
        """Upper bound of variables."""
        return self._upper_bound_variable

    @property
    def c_lb(self) -> NDArray[np.float64]:
        """Lower bound of constraints."""
        return self._lower_bound_constraint

    @property
    def c_ub(self) -> NDArray[np.float64]:
        """Upper bound of constraints."""
        return self._upper_bound_constraint

    @property
    def N(self) -> int:
        """Number of phases."""
        return self.n_p

    @property
    def L(self) -> int:
        return self.r_s

    @property
    def ok(self) -> bool:
        return self._phase_set and self._objective_set and self._system_constraint_set

    def _basic_value(self, x):
        sp = x[self.l_s:self.r_s]
        V = [x[l:r] for l, r in zip(self.l_p, self.r_p)]
        SP = [np.repeat(sp, p.l0_m) for p in self.p]
        MT = [(v[-1] + v[-2]) / 2 for v in V]
        DT = [v[-1] - v[-2] for v in V]
        T = [(p.t_m - 0.5) * dt + mt for p, v, mt, dt in zip(self.p, V, MT, DT)]
        M = [np.concatenate((p.f_v2m(v[:-2]), s, t)) for p, v, s, t in zip(self.p, V, SP, T)]
        I = [f_I.F(m, p.l0_m) @ p.w * dt for p, m, dt in zip(self.p, M, DT) for f_I in p.F_I]
        v_s = np.concatenate([sp, I])
        return sp, V, SP, MT, DT, T, M, I, v_s

    def _basic_gradient_static_parameter(self, sp_, suffix=''):
        return '', [('np.array([1.], dtype=np.float64)', np.array([self.l_s + sp_], dtype=np.int32))]

    def _basic_gradient_integral(self, p_, I_, suffix=''):
        p = self.p[p_]
        init = ''
        grad = []
        init += f'p_gi{suffix} = self.p[{p_}]\n'
        init += f'f_gi{suffix} = p_gi{suffix}.F_I[{I_}].F(M[{p_}], p_gi{suffix}.l0_m) @ p_gi{suffix}.w\n'
        init += f'g_gi{suffix} = p_gi{suffix}.F_I[{I_}].G(M[{p_}], p_gi{suffix}.l0_m) * p_gi{suffix}.w * DT[{p_}]\n'
        for row, n_ in enumerate(p.F_I[I_].G_index):
            if n_ < p.n:
                if p.c[n_]:
                    grad.append(
                        (f'p_gi{suffix}.f_nc2c(g_gi{suffix}[{row}])', np.arange(p.l_v[n_], p.r_v[n_]) + self.l_p[p_]))
                else:
                    grad.append((f'g_gi{suffix}[{row}]', np.arange(p.l_v[n_], p.r_v[n_]) + self.l_p[p_]))
            elif n_ < p.n + self.n_s:
                grad.append(
                    (f'np.array([np.sum(g_gi{suffix}[{row}])], dtype=np.float64)', np.array([self.l_s + n_ - p.n])))
            else:
                grad.append((f'np.array([g_gi{suffix}[{row}] @ (1 - p_gi{suffix}.t_m),'
                             f' g_gi{suffix}[{row}] @ p_gi{suffix}.t_m])',
                             np.array([self.r_p[p_] - 2, self.r_p[p_] - 1])))
        grad.append((f'np.array([-f_gi{suffix}, f_gi{suffix}], dtype=np.float64)',
                     np.array([self.r_p[p_] - 2, self.r_p[p_] - 1])))
        return init, grad

    def _basic_gradient_phase_func(self, p_: int, func_name: str, suffix=''):
        p = self.p[p_]
        init = ''
        grad = []
        init += f'p_gpf{suffix} = self.p[{p_}]\n'
        init += f'g_gpf{suffix} = p_gpf{suffix}.{func_name}.G(M[{p_}], p_gpf{suffix}.l0_m)\n'
        func = eval(f'p.{func_name}')
        for row, n_ in enumerate(func.G_index):
            if n_ < p.n:
                if p.c[n_]:
                    v_index = np.concatenate(
                        [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[n_] + self.l_p[p_] for i in range(p.N)])
                    grad.append((f'g_gpf{suffix}[{row}]', v_index))
                else:
                    grad.append((f'g_gpf{suffix}[{row}]', np.arange(p.l_v[n_], p.r_v[n_]) + self.l_p[p_]))
            elif n_ < p.n + self.n_s:
                grad.append((f'g_gpf{suffix}[{row}]',
                             np.full(p.l0_m, self.l_s + n_ - p.n)))
            else:
                grad.append(
                    (f'(g_gpf{suffix}[{row}] * (1 - p_gpf{suffix}.t_m))', np.full(p.l0_m, self.r_p[p_] - 2)))
                grad.append((f'(g_gpf{suffix}[{row}] * p_gpf{suffix}.t_m)', np.full(p.l0_m, self.r_p[p_] - 1)))
        return init, grad

    def _basic_gradient_system(self, func_name: str, suffix=''):
        init = f'g_gs{suffix} = self.{func_name}.G(v_s, 1)\n'
        grad = []
        func = eval('self.' + func_name)
        for row, n_ in enumerate(func.G_index):
            if n_ < self.n_s:
                init_, grad_ = self._basic_gradient_static_parameter(n_, suffix=f'_gs_sp{n_}{suffix}')
                init += init_
                grad.extend(map(lambda c, i: (c + f' * g_gs{suffix}[{row}]', i), *zip(*grad_)))
            else:
                p_, I_ = self._mapping_integral[n_ - self.n_s]
                init_, grad_ = self._basic_gradient_integral(p_, I_, suffix=f'_gs_i{p_}{I_}{suffix}')
                init += init_
                grad.extend(map(lambda c, i: (c + f' * g_gs{suffix}[{row}]', i), *zip(*grad_)))
        return init, grad

    def _basic_hessian_integral(self, p_, I_, suffix=''):
        p = self.p[p_]
        init = ''
        hess = []  # code, row, col (row >= col)
        init += f'p_hi{suffix} = self.p[{p_}]\n'
        init += f'g_hi{suffix} = p_hi{suffix}.F_I[{I_}].G(M[{p_}], p_hi{suffix}.l0_m) * p_hi{suffix}.w\n'
        init += f'h_hi{suffix} = p_hi{suffix}.F_I[{I_}].H(M[{p_}], p_hi{suffix}.l0_m) * p_hi{suffix}.w * DT[{p_}]\n'
        # 2 parts: v/v, dt/v (dt/dt = 0)
        # v/v
        for row, (x_, y_) in enumerate(zip(*p.F_I[I_].H_index)):  # x_ >= y_
            if x_ < p.n:
                if p.c[x_]:
                    x_index = np.concatenate(
                        [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[x_] + self.l_p[p_] for i in range(p.N)])
                else:
                    x_index = np.arange(p.l_v[x_], p.r_v[x_]) + self.l_p[p_]
                if p.c[y_]:
                    y_index = np.concatenate(
                        [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[y_] + self.l_p[p_] for i in range(p.N)])
                else:
                    y_index = np.arange(p.l_v[y_], p.r_v[y_]) + self.l_p[p_]
                hess.append((f'h_hi{suffix}[{row}]', x_index, y_index))
            elif x_ < p.n + self.n_s:
                if y_ < p.n:
                    if p.c[y_]:
                        y_index = np.concatenate(
                            [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[y_] + self.l_p[p_] for i in range(p.N)])
                    else:
                        y_index = np.arange(p.l_v[y_], p.r_v[y_]) + self.l_p[p_]
                    hess.append((f'h_hi{suffix}[{row}]', np.full(p.l0_m, self.l_s + x_ - p.n), y_index))
                else:
                    hess.append((f'np.array([np.sum(h_hi{suffix}[{row}])], dtype=np.float64)',
                                 np.array([self.l_s + x_ - p.n]), np.array([self.l_s + y_ - p.n])))
            else:
                if y_ < p.n:
                    if p.c[y_]:
                        y_index = np.concatenate(
                            [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[y_] + self.l_p[p_] for i in range(p.N)])
                    else:
                        y_index = np.arange(p.l_v[y_], p.r_v[y_]) + self.l_p[p_]
                    hess.append(
                        (f'h_hi{suffix}[{row}] * (1 - p_hi{suffix}.t_m)', np.full(p.l0_m, self.r_p[p_] - 2), y_index))
                    hess.append((f'h_hi{suffix}[{row}] * p_hi{suffix}.t_m', np.full(p.l0_m, self.r_p[p_] - 1), y_index))
                elif y_ < p.n + self.n_s:
                    hess.append((f'np.array([np.sum(h_hi{suffix}[{row}] * (1 - p_hi{suffix}.t_m))], dtype=np.float64)',
                                 np.array([self.l_s + y_ - p.n]), np.array([self.r_p[p_] - 2])))
                    hess.append((f'np.array([np.sum(h_hi{suffix}[{row}] * p_hi{suffix}.t_m)], dtype=np.float64)',
                                 np.array([self.l_s + y_ - p.n]), np.array([self.r_p[p_] - 1])))
                else:
                    hess.append((
                        f'np.array([np.sum(h_hi{suffix}[{row}] * (1 - p_hi{suffix}.t_m) * (1 - p_hi{suffix}.t_m))], dtype=np.float64)',
                        np.array([self.r_p[p_] - 2]), np.array([self.r_p[p_] - 2])))
                    hess.append((
                        f'np.array([np.sum(h_hi{suffix}[{row}] * p_hi{suffix}.t_m * (1 - p_hi{suffix}.t_m))], dtype=np.float64)',
                        np.array([self.r_p[p_] - 1]), np.array([self.r_p[p_] - 2])))
                    hess.append((
                        f'np.array([np.sum(h_hi{suffix}[{row}] * p_hi{suffix}.t_m * p_hi{suffix}.t_m)], dtype=np.float64)',
                        np.array([self.r_p[p_] - 1]), np.array([self.r_p[p_] - 1])))
        # t/v
        for row, y_ in enumerate(p.F_I[I_].G_index):  # x_ >= y_
            if y_ < p.n:
                if p.c[y_]:
                    y_index = np.concatenate(
                        [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[y_] + self.l_p[p_] for i in range(p.N)])
                else:
                    y_index = np.arange(p.l_v[y_], p.r_v[y_]) + self.l_p[p_]
                hess.append((f'-g_hi{suffix}[{row}]', np.full(p.l0_m, self.r_p[p_] - 2), y_index))
                hess.append((f'g_hi{suffix}[{row}]', np.full(p.l0_m, self.r_p[p_] - 1), y_index))
            elif y_ < p.n + self.n_s:
                hess.append(
                    (f'np.array([-np.sum(g_hi{suffix}[{row}])], dtype=np.float64)', np.array([self.l_s + y_ - p.n]),
                     np.array([self.r_p[p_] - 2])))
                hess.append(
                    (f'np.array([np.sum(g_hi{suffix}[{row}])], dtype=np.float64)', np.array([self.l_s + y_ - p.n]),
                     np.array([self.r_p[p_] - 1])))
            else:
                hess.append((f'np.array([2 * -np.sum(g_hi{suffix}[{row}] * (1 - p_hi{suffix}.t_m))], dtype=np.float64)',
                             np.array([self.r_p[p_] - 2]), np.array([self.r_p[p_] - 2])))
                hess.append((f'np.array([np.sum(g_hi{suffix}[{row}] * (1 - p_hi{suffix}.t_m * 2))], dtype=np.float64)',
                             np.array([self.r_p[p_] - 1]), np.array([self.r_p[p_] - 2])))
                hess.append((f'np.array([2 * np.sum(g_hi{suffix}[{row}] * p_hi{suffix}.t_m)], dtype=np.float64)',
                             np.array([self.r_p[p_] - 1]), np.array([self.r_p[p_] - 1])))
        return init, hess

    def _basic_hessian_phase_func(self, p_: int, func_name: str, suffix=''):
        p = self.p[p_]
        init = ''
        hess = []  # code, row, col (row >= col)
        init += f'p_hpf{suffix} = self.p[{p_}]\n'
        init += f'g_hpf{suffix} = p_hpf{suffix}.{func_name}.G(M[{p_}], p_hpf{suffix}.l0_m)\n'
        init += f'h_hpf{suffix} = p_hpf{suffix}.{func_name}.H(M[{p_}], p_hpf{suffix}.l0_m)\n'
        func = eval('p.' + func_name)
        for row, (x_, y_) in enumerate(zip(*func.H_index)):  # x_ >= y_
            if x_ < p.n:
                if p.c[x_]:
                    x_index = np.concatenate(
                        [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[x_] + self.l_p[p_] for i in range(p.N)])
                else:
                    x_index = np.arange(p.l_v[x_], p.r_v[x_]) + self.l_p[p_]
                if p.c[y_]:
                    y_index = np.concatenate(
                        [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[y_] + self.l_p[p_] for i in range(p.N)])
                else:
                    y_index = np.arange(p.l_v[y_], p.r_v[y_]) + self.l_p[p_]
                hess.append((f'h_hpf{suffix}[{row}]', x_index, y_index))
            elif x_ < p.n + self.n_s:
                if y_ < p.n:
                    if p.c[y_]:
                        y_index = np.concatenate(
                            [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[y_] + self.l_p[p_] for i in range(p.N)])
                    else:
                        y_index = np.arange(p.l_v[y_], p.r_v[y_]) + self.l_p[p_]
                    hess.append((f'h_hpf{suffix}[{row}]', np.full(p.l0_m, self.l_s + x_ - p.n), y_index))
                else:
                    hess.append((f'h_hpf{suffix}[{row}]', np.full(p.l0_m, self.l_s + x_ - p.n),
                                 np.full(p.l0_m, self.l_s + y_ - p.n)))
            else:
                if y_ < p.n:
                    if p.c[y_]:
                        y_index = np.concatenate(
                            [np.arange(p.l_c[i], p.r_c[i]) + p.l_v[y_] + self.l_p[p_] for i in range(p.N)])
                    else:
                        y_index = np.arange(p.l_v[y_], p.r_v[y_]) + self.l_p[p_]
                    hess.append(
                        (f'h_hpf{suffix}[{row}] * (1 - p_hpf{suffix}.t_m)', np.full(p.l0_m, self.r_p[p_] - 2), y_index))
                    hess.append(
                        (f'h_hpf{suffix}[{row}] * p_hpf{suffix}.t_m', np.full(p.l0_m, self.r_p[p_] - 1), y_index))
                elif y_ < p.n + self.n_s:
                    hess.append((f'h_hpf{suffix}[{row}] * (1 - p_hpf{suffix}.t_m)',
                                 np.full(p.l0_m, self.l_s + y_ - p.n), np.full(p.l0_m, self.r_p[p_] - 2)))
                    hess.append((f'h_hpf{suffix}[{row}] * p_hpf{suffix}.t_m',
                                 np.full(p.l0_m, self.l_s + y_ - p.n), np.full(p.l0_m, self.r_p[p_] - 1)))
                else:
                    hess.append((f'h_hpf{suffix}[{row}] * (1 - p_hpf{suffix}.t_m) * (1 - p_hpf{suffix}.t_m)',
                                 np.full(p.l0_m, self.r_p[p_] - 2), np.full(p.l0_m, self.r_p[p_] - 2)))
                    hess.append((f'h_hpf{suffix}[{row}] * p_hpf{suffix}.t_m * (1 - p_hpf{suffix}.t_m)',
                                 np.full(p.l0_m, self.r_p[p_] - 1), np.full(p.l0_m, self.r_p[p_] - 2)))
                    hess.append((f'h_hpf{suffix}[{row}] * p_hpf{suffix}.t_m * p_hpf{suffix}.t_m',
                                 np.full(p.l0_m, self.r_p[p_] - 1), np.full(p.l0_m, self.r_p[p_] - 1)))
        return init, hess

    def _basic_hessian_system(self, func_name: str, suffix=''):
        init = (f'g_hs{suffix} = self.{func_name}.G(v_s, 1)\n'
                f'h_hs{suffix} = self.{func_name}.H(v_s, 1)\n')
        hess = []
        func = eval('self.' + func_name)
        for row, n_ in enumerate(func.G_index):
            if n_ >= self.n_s:
                p_, I_ = self._mapping_integral[n_ - self.n_s]
                init_, hess_ = self._basic_hessian_integral(p_, I_, suffix=f'_hs_i{p_}{I_}{suffix}')
                init += init_
                for code, x_, y_ in hess_:
                    hess.append((code + f' * g_hs{suffix}[{row}]', x_, y_))

        for row, (x_, y_) in enumerate(zip(*func.H_index)):
            if x_ < self.n_s:
                init_x, grad_x = self._basic_gradient_static_parameter(x_, suffix=f'_hs_sp{y_}{suffix}')
            else:
                p_, I_ = self._mapping_integral[x_ - self.n_s]
                init_x, grad_x = self._basic_gradient_integral(p_, I_, suffix=f'_hs_i{p_}{I_}{suffix}')
            if y_ < self.n_s:
                init_y, grad_y = self._basic_gradient_static_parameter(y_, suffix=f'_hs_sp{y_}{suffix}')
            else:
                p_, I_ = self._mapping_integral[y_ - self.n_s]
                init_y, grad_y = self._basic_gradient_integral(p_, I_, suffix=f'_hs_i{p_}{I_}{suffix}')
            init += init_x + init_y

            for c_x, i_x in grad_x:
                for c_y, i_y in grad_y:
                    row_ = np.repeat(i_x, len(i_y))
                    col_ = np.tile(i_y, len(i_x))
                    if x_ == y_:
                        index, row2_, col2_ = _filter_rc(row_, col_)
                    else:
                        index, row2_, col2_ = _rotate_rc(row_, col_)
                    if np.array_equal(index, np.arange(len(i_x) * len(i_y))):
                        hess.append((f'np.kron({c_x}, {c_y}) * h_hs{suffix}[{row}]', row2_, col2_))
                    elif len(index) > 0:
                        index_str = 'np.array([' + ', '.join(map(str, index)) + '], dtype=np.int32)'
                        hess.append((f'np.kron({c_x}, {c_y})[{index_str}] * h_hs{suffix}[{row}]', row2_, col2_))
        return init, hess

    def objective(self, x):
        v_s = self._basic_value(x)[-1]
        return self.F_o.F(v_s, 1)[0]

    def gradient(self, x):
        Grad = np.zeros_like(x)
        sp, V, SP, MT, DT, T, M, I, v_s = self._basic_value(x)
        init, grad = self._basic_gradient_system('F_o')
        exec(init, globals(), locals())
        for c, i in grad:
            Grad[i] += eval(c)
        return Grad

    def constraints(self, x):
        sp, V, SP, MT, DT, T, M, I, v_s = self._basic_value(x)

        c_system = np.array([f_c.F(v_s, 1)[0] for f_c in self.F_c], dtype=np.float64)

        d_est_phase = np.array([p.D_x @ v[:p.L_x] / dt for p, v, dt in zip(self.p, V, DT)], dtype=np.float64).flatten()
        d_func_phase = np.array([f_d.F(m, p.l0_m) for p, m in zip(self.p, M) for f_d in p.F_d],
                                dtype=np.float64).flatten()
        c_dynamics = d_est_phase - d_func_phase

        c_phase = np.array([f_c.F(m, p.l0_m) for p, m in zip(self.p, M) for f_c in p.F_c], dtype=np.float64).flatten()

        c_boundary = np.array([v_bc - f_bc.F(sp, 1)[0] for p, v in zip(self.p, V)
                               for f_bc, v_bc in zip(p.F_bc, v[p.i_bc])], dtype=np.float64)

        return np.concatenate([c_system, c_dynamics, c_phase, c_boundary])

    def _update_jacobian_structure(self):
        L_c_system = self.n_c
        L_c_dynamics = sum(p.n_x * p.l0_m for p in self.p)
        L_c_phase = sum(p.n_c * p.l0_m for p in self.p)
        L_c_boundary = sum(p.n_c for p in self.p)
        l_c = np.cumsum([0, L_c_system, L_c_dynamics, L_c_phase])

        Jac_row = []
        Jac_col = []
        Jac_code = 'jac_data = np.zeros({})\n'

        # system constraint
        for c_ in range(self.n_c):
            init, grad = self._basic_gradient_system(f'F_c[{c_}]', suffix=f'_c{c_}')
            Jac_code += init
            for c, i in grad:
                Jac_code += f'jac_data[{len(Jac_row)}:{len(Jac_row) + len(i)}] = {c}\n'
                Jac_row.extend(np.full_like(i, l_c[0] + c_))
                Jac_col.extend(i)

        c_base_ = l_c[1]
        for p_, p in enumerate(self._phase):
            # dynamics constraint: est
            Jac_code += f'p = self.p[{p_}]\n'
            Jac_code += f'jac_data[{len(Jac_row)}:{len(Jac_row) + p.D_coo.nnz}] = p.D_coo.data / DT[{p_}]\n'
            Jac_row += [c_base_ + r for r in p.D_coo.row]
            Jac_col += [self.l_p[p_] + c for c in p.D_coo.col]

            Jac_code += f'grad_t = -(p.D_x @ V[{p_}][:p.L_x]) / DT[{p_}] / DT[{p_}]\n'
            Jac_code += f'jac_data[{len(Jac_row)}:{len(Jac_row) + p.n_x * p.l0_m}] = -grad_t\n'
            Jac_row += [c_base_ + r for r in range(p.n_x * p.l0_m)]
            Jac_col += [self.r_p[p_] - 2] * p.n_x * p.l0_m
            Jac_code += f'jac_data[{len(Jac_row)}:{len(Jac_row) + p.n_x * p.l0_m}] = grad_t\n'
            Jac_row += [c_base_ + r for r in range(p.n_x * p.l0_m)]
            Jac_col += [self.r_p[p_] - 1] * p.n_x * p.l0_m

            # dynamics constraint: func
            for d_, f_d in enumerate(p.F_d):
                init, grad = self._basic_gradient_phase_func(p_, f'F_d[{d_}]', suffix=f'_p{p_}d{d_}')
                Jac_code += init
                for c, i in grad:
                    Jac_code += f'jac_data[{len(Jac_row)}:{len(Jac_row) + p.l0_m}] = -{c}\n'
                    Jac_row.extend(np.arange(c_base_ + d_ * p.l0_m, c_base_ + (d_ + 1) * p.l0_m))
                    Jac_col.extend(i)

            c_base_ += p.n_x * p.l0_m

        # phase constraint
        c_base_ = l_c[2]
        for p_, p in enumerate(self._phase):
            for c_, f_c in enumerate(p.F_c):
                init, grad = self._basic_gradient_phase_func(p_, f'F_c[{c_}]', suffix=f'_p{p_}c{c_}')
                Jac_code += init
                for c, i in grad:
                    Jac_code += f'jac_data[{len(Jac_row)}:{len(Jac_row) + p.l0_m}] = {c}\n'
                    Jac_row.extend(np.arange(c_base_ + c_ * p.l0_m, c_base_ + (c_ + 1) * p.l0_m))
                    Jac_col.extend(i)
            c_base_ += p.n_c * p.l0_m

        # boundary constraint
        c_base_ = l_c[3]
        for p_, p in enumerate(self._phase):
            if p.n_bc > 0:
                Jac_code += f'p = self.p[{p_}]\n'
                Jac_code += f'jac_data[{len(Jac_row)}:{len(Jac_row) + p.n_bc}] = np.ones({p.n_bc}, dtype=np.float64)\n'
                Jac_row.extend(range(c_base_, c_base_ + p.n_bc))
                Jac_col.extend(self.l_p[p_] + p.i_bc)
            for bc_, bc in enumerate(p.F_bc):
                Jac_code += f'g_p = p.F_bc[{bc_}].G(sp, 1)\n'
                for row, v_ in enumerate(bc.G_index):
                    Jac_code += f'jac_data[{len(Jac_row)}] = -g_p[{row}]\n'
                    Jac_row.append(c_base_ + bc_)
                    Jac_col.append(self.l_s + v_)
            c_base_ += p.n_bc

        self._jac_row = np.array(Jac_row, dtype=np.int32)
        self._jac_col = np.array(Jac_col, dtype=np.int32)
        self._jac_func_code = Jac_code.format(len(Jac_row))

    def jacobian(self, x):
        sp, V, SP, MT, DT, T, M, I, v_s = self._basic_value(x)
        exec(self._jac_func_code, globals(), locals())
        return locals()['jac_data']

    def jacobianstructure(self):
        return self._jac_row, self._jac_col

    def _update_hessian_structure(self):
        L_c_system = self.n_c
        L_c_dynamics = sum(p.n_x * p.l0_m for p in self.p)
        L_c_phase = sum(p.n_c * p.l0_m for p in self.p)
        L_c_boundary = sum(p.n_c for p in self.p)
        l_c = np.cumsum([0, L_c_system, L_c_dynamics, L_c_phase])

        Hess_row = []
        Hess_col = []
        Hess_code = 'hess_data = np.zeros({})\n'

        # objective
        init_, hess_ = self._basic_hessian_system('F_o')
        Hess_code += init_
        for code, row, col in hess_:
            Hess_code += f'hess_data[{len(Hess_row)}:{len(Hess_row) + len(row)}] = {code} * fct_o\n'
            Hess_row.extend(row)
            Hess_col.extend(col)

        # system constraint
        for c_ in range(self.n_c):
            init_, hess_ = self._basic_hessian_system(f'F_c[{c_}]')
            Hess_code += init_
            for code, row, col in hess_:
                Hess_code += f'hess_data[{len(Hess_row)}:{len(Hess_row) + len(row)}] = {code} * fct_c[{c_}]\n'
                Hess_row.extend(row)
                Hess_col.extend(col)

        # dynamics constraint
        c_base_ = l_c[1]
        for p_, p in enumerate(self._phase):
            Hess_code += f'p = self.p[{p_}]\n'
            Hess_code += f'hess_data[{len(Hess_row)}:{len(Hess_row) + p.D_coo.nnz}] ' \
                         f'= p.D_coo.data / DT[{p_}] / DT[{p_}] * fct_c[{c_base_} + p.D_coo.row]\n'
            Hess_row.extend(np.full_like(p.D_coo.col, self.r_p[p_] - 2))
            Hess_col.extend(self.l_p[p_] + p.D_coo.col)
            Hess_code += f'hess_data[{len(Hess_row)}:{len(Hess_row) + p.D_coo.nnz}] ' \
                         f'= -p.D_coo.data / DT[{p_}] / DT[{p_}] * fct_c[{c_base_} + p.D_coo.row]\n'
            Hess_row.extend(np.full_like(p.D_coo.col, self.r_p[p_] - 1))
            Hess_col.extend(self.l_p[p_] + p.D_coo.col)

            Hess_code += f'hess_t = 2 * (p.D_x @ V[{p_}][:p.L_x]) / (DT[{p_}] ** 3) ' \
                         f'* fct_c[{c_base_}:{c_base_ + p.n_x * p.l0_m}]\n'
            Hess_code += f'hess_data[{len(Hess_row)}:{len(Hess_row) + p.n_x * p.l0_m}] = hess_t\n'
            Hess_row.extend(np.full(p.n_x * p.l0_m, self.r_p[p_] - 2))
            Hess_col.extend(np.full(p.n_x * p.l0_m, self.r_p[p_] - 2))
            Hess_code += f'hess_data[{len(Hess_row)}:{len(Hess_row) + p.n_x * p.l0_m}] = -hess_t\n'
            Hess_row.extend(np.full(p.n_x * p.l0_m, self.r_p[p_] - 1))
            Hess_col.extend(np.full(p.n_x * p.l0_m, self.r_p[p_] - 2))
            Hess_code += f'hess_data[{len(Hess_row)}:{len(Hess_row) + p.n_x * p.l0_m}] = hess_t\n'
            Hess_row.extend(np.full(p.n_x * p.l0_m, self.r_p[p_] - 1))
            Hess_col.extend(np.full(p.n_x * p.l0_m, self.r_p[p_] - 1))

            for d_ in range(p.n_x):
                init, hess = self._basic_hessian_phase_func(p_, f'F_d[{d_}]')
                Hess_code += init
                for code, row, col in hess:
                    Hess_code += f'hess_data[{len(Hess_row)}:{len(Hess_row) + len(row)}] ' \
                                 f'= -{code} * fct_c[{c_base_ + d_ * p.l0_m}:{c_base_ + (d_ + 1) * p.l0_m}]\n'
                    Hess_row.extend(row)
                    Hess_col.extend(col)

            c_base_ += p.n_x * p.l0_m

        # phase constraint
        c_base_ = l_c[2]
        for p_ in range(self.n_p):
            p = self.p[p_]
            for c_ in range(self.p[p_].n_c):
                init, hess = self._basic_hessian_phase_func(p_, f'F_c[{c_}]')
                Hess_code += init
                for code, row, col in hess:
                    Hess_code += f'hess_data[{len(Hess_row)}:{len(Hess_row) + len(row)}] ' \
                                 f'= {code} * fct_c[{c_base_ + c_ * p.l0_m}:{c_base_ + (c_ + 1) * p.l0_m}]\n'
                    Hess_row.extend(row)
                    Hess_col.extend(col)
            c_base_ += p.n_c * p.l0_m

        # boundary constraint
        c_base_ = l_c[3]
        for p_, p in enumerate(self._phase):
            if p.n_bc > 0:
                Hess_code += f'p = self.p[{p_}]\n'
            for bc_, bc in enumerate(p.F_bc):
                Hess_code += f'h_p = p.F_bc[{bc_}].H(sp, 1)\n'
                for row, (x_, y_) in enumerate(zip(*bc.H_index)):
                    Hess_code += f'hess_data[{len(Hess_row)}] = -h_p[{row}] * fct_c[{c_base_ + bc_}]\n'
                    Hess_row.append(self.l_s + x_)
                    Hess_col.append(self.l_s + y_)
            c_base_ += p.n_bc

        self._hess_row = np.array(Hess_row, dtype=np.int32)
        self._hess_col = np.array(Hess_col, dtype=np.int32)
        self._hess_func_code = Hess_code.format(len(Hess_row))

    def hessianstructure(self):
        return self._hess_row, self._hess_col

    def hessian(self, x, fct_c, fct_o):
        sp, V, SP, MT, DT, T, M, I, v_s = self._basic_value(x)
        exec(self._hess_func_code, globals(), locals())
        return locals()['hess_data']


def _filter_rc(row, col):
    row2 = []
    col2 = []
    index = []
    for i, (r, c) in enumerate(zip(row, col)):
        if r >= c:
            row2.append(r)
            col2.append(c)
            index.append(i)
    return np.array(index, dtype=np.int32), np.array(row2, dtype=np.int32), np.array(col2, dtype=np.int32)


def _rotate_rc(row, col):
    row2 = []
    col2 = []
    index = []
    for i in range(len(row)):
        if row[i] > col[i]:
            row2.append(row[i])
            col2.append(col[i])
            index.append(i)
        elif row[i] == col[i]:
            row2.append(row[i])
            col2.append(col[i])
            index.append(i)
            row2.append(row[i])
            col2.append(col[i])
            index.append(i)
        else:
            row2.append(col[i])
            col2.append(row[i])
            index.append(i)
    return np.array(index, dtype=np.int32), np.array(row2, dtype=np.int32), np.array(col2, dtype=np.int32)



import functools
import os
import tempfile
from typing import List, Tuple, Union

import numpy as np
import sympy as sp
from numpy.typing import NDArray
from sympy.codegen.rewriting import create_expand_pow_optimization
from sympy.printing.lambdarepr import lambdarepr


@functools.lru_cache
def _deco_basic(num_free_symbol: int, parallel: bool, fastmath: bool) -> str:
    """Decorator for basic functions."""
    tail = ''
    # maybe numba bugs, do not compile if set
    # if parallel:
    #     tail += ', target="parallel"'
    if fastmath:
        tail += ', fastmath=True'
    return '@numba.vectorize("float64({})"{})\n'.format(', '.join(['float64'] * num_free_symbol), tail)


@functools.lru_cache
def _deco_F(parallel: bool, fastmath: bool) -> str:
    """Decorator for aggregate function F (value)."""
    tail = ''
    if parallel:
        tail += ', parallel=True'
    if fastmath:
        tail += ', fastmath=True'
    return '@numba.njit("float64[:](float64[:], int32)"{})\n'.format(tail)


@functools.lru_cache
def _deco_GH(parallel: bool, fastmath: bool) -> str:
    """Decorator for aggregate functions G (gradient) & H (hessian)."""
    tail = ''
    if parallel:
        tail += ', parallel=True'
    if fastmath:
        tail += ', fastmath=True'
    return '@numba.njit("float64[:, :](float64[:], int32)"{})\n'.format(tail)


class FastFunc:
    """FastFunc take a sympy expression and a list of sympy symbols as function arguments,
    and generate JITed, vectorized functions for value, gradient and hessian.
    Gradient and hessian are generated by automatic differentiation and in sparse format,
    i.e., only non-zero elements are computed. The index of non-zero elements are stored in
    G_index and H_index, respectively.

    Both F, G, and H take a second argument for length of a single variable, the allows the
    user change the discretization scheme without recompiling the functions.
    """

    def __init__(self, function: Union[int, float, sp.Expr], args: List[sp.Symbol], /,
                 simplify: bool = False, parallel: bool = False, fastmath: bool = False) -> None:
        self._function = sp.sympify(function)
        self._args = args

        # random valid python name to avoid conflict
        self._valid_args = [sp.Symbol('xbN4dRhrnN_{}'.format(i)) for i in range(len(args))]
        for i in range(len(args)):
            self._function = self._function.subs(args[i], self._valid_args[i])
        self._args = self._valid_args

        self._simplify = simplify
        self._parallel = parallel
        self._fastmath = fastmath

        self._value_index = []
        self._grad_index = []
        self._hess_index = []

        self._value_consts = None
        self._grad_consts = {}
        self._hess_consts = {}

        self.__expand_opt = create_expand_pow_optimization(3)

        basic_str = self._gen_basic_funcs()
        aggregate_str = self._gen_aggregate_funcs()

        self._compile(basic_str + aggregate_str)
        self.G_index = self._gen_G_index()
        self.H_index = self._gen_H_index()

    def _compile(self, code: str) -> None:
        tmp_file = tempfile.NamedTemporaryFile(mode='w', suffix='.py', delete=False)

        tmp_file.write('from numpy import *\n')
        tmp_file.write('import numba\n')

        tmp_file.write(code)
        tmp_file.close()

        import importlib.util
        import sys

        spec = importlib.util.spec_from_file_location('tmp_module', tmp_file.name)
        tmp_module = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = tmp_module
        spec.loader.exec_module(tmp_module)

        self.F = tmp_module.F
        self.G = tmp_module.G
        self.H = tmp_module.H

        os.unlink(tmp_file.name)

    def _free_symbols(self, expr: sp.Expr) -> List[int]:
        return sorted([self._args.index(sym) for sym in expr.free_symbols])

    def _gen_basic_funcs(self) -> str:
        result_str = ''
        func = self._function

        # value
        if self._simplify:
            func = sp.simplify(func)
        if func.is_constant():
            func = sp.sympify(func)
            self._value_consts = float(func)
            return ''

        free_symbols_value = self._free_symbols(func)
        self._value_index = free_symbols_value
        deco_str = _deco_basic(len(free_symbols_value), self._parallel, self._fastmath)
        func_str = lambdarepr(self.__expand_opt(func)).replace('math.', '')
        param_str = ', '.join([self._args[i].name for i in free_symbols_value])
        result_str += deco_str + 'def f({}):\n'.format(param_str) + '    return ' + func_str + '\n\n'

        for j in free_symbols_value:  # row in hess
            # gradient
            grad = sp.diff(func, self._args[j])
            if self._simplify:
                grad = sp.simplify(grad)
            if grad == 0:
                continue
            if grad.is_constant():
                grad = sp.simplify(grad)
                self._grad_consts[j] = float(grad)
                self._grad_index.append((j, []))
                continue

            free_symbols_grad = self._free_symbols(grad)
            self._grad_index.append((j, free_symbols_grad))
            deco_str = _deco_basic(len(free_symbols_grad), self._parallel, self._fastmath)
            grad_str = lambdarepr(self.__expand_opt(grad)).replace('math.', '')
            param_str = ', '.join([self._args[k].name for k in free_symbols_grad])
            result_str += deco_str + 'def df_d{}({}):\n'.format(j, param_str) \
                          + '    return ' + grad_str + '\n\n'

            # hessian
            for k in free_symbols_value:  # col
                if j < k:
                    break
                hess = sp.diff(grad, self._args[k])
                if self._simplify:
                    hess = sp.simplify(hess)
                if hess == 0:
                    continue
                if hess.is_constant():
                    hess = sp.simplify(hess)
                    self._hess_consts[(j, k)] = float(hess)
                    self._hess_index.append((j, k, []))
                    continue

                free_symbols_hess = self._free_symbols(hess)
                self._hess_index.append((j, k, free_symbols_hess))
                deco_str = _deco_basic(len(free_symbols_hess), self._parallel, self._fastmath)
                hess_str = lambdarepr(self.__expand_opt(hess)).replace('math.', '')
                param_str = ', '.join([self._args[l].name for l in free_symbols_hess])
                result_str += deco_str + 'def d2f_d{}d{}({}):\n'.format(j, k, param_str) \
                              + '    return ' + hess_str + '\n\n'
        return result_str

    def _gen_aggregate_funcs(self) -> str:
        result_str = ''
        # value
        result_str += _deco_F(self._parallel, self._fastmath)
        result_str += 'def F(x, n):\n'
        result_str += '    r = zeros(n, dtype=float64)\n'
        if self._value_consts is not None:
            result_str += f'    r[:] = {self._value_consts}\n'
        else:
            param_str = ', '.join([f'x[{j} * n:{j + 1} * n]' for j in self._value_index])
            result_str += '    r = f({})\n'.format(param_str)
        result_str += '    return r\n\n'

        # gradient
        result_str += _deco_GH(self._parallel, self._fastmath)
        result_str += 'def G(x, n):\n'
        result_str += f'    r = zeros(({len(self._grad_index)}, n), dtype=float64)\n'
        for row, (x, fs) in enumerate(self._grad_index):  # fs for free symbols
            if x in self._grad_consts:
                result_str += f'    r[{row}] = {self._grad_consts[x]}\n'
            else:
                param_str = ', '.join([f'x[{j} * n:{j + 1} * n]' for j in fs])
                result_str += '    r[{}] = df_d{}({})\n'.format(row, x, param_str)
        result_str += '    return r\n\n'

        # hessian
        result_str += _deco_GH(self._parallel, self._fastmath)
        result_str += 'def H(x, n):\n'
        result_str += f'    r = zeros(({len(self._hess_index)}, n), dtype=float64)\n'
        for row, (x, y, fs) in enumerate(self._hess_index):
            if (x, y) in self._hess_consts:
                result_str += f'    r[{row}] = {self._hess_consts[(x, y)]}\n'
            else:
                param_str = ', '.join([f'x[{j} * n:{j + 1} * n]' for j in fs])
                result_str += '    r[{}] = d2f_d{}d{}({})\n'.format(row, x, y, param_str)
        result_str += '    return r\n\n'

        return result_str

    def _gen_G_index(self) -> NDArray[np.int32]:
        return np.array([x for x, _ in self._grad_index], dtype=np.int32)

    def _gen_H_index(self) -> Tuple[NDArray[np.int32], NDArray[np.int32]]:
        return np.array([x for x, _, _ in self._hess_index], dtype=np.int32), \
               np.array([y for _, y, _ in self._hess_index], dtype=np.int32)



from typing import Union, List, Iterable

import sympy as sp

from .discretization import *
from .fastfunc import FastFunc


class Phase:
    def __init__(self, state: Union[int, List[str]], control: Union[int, List[str]],
                 symbol_static_parameter: List[sp.Symbol], identifier: int, /, simplify: bool = False,
                 parallel: bool = False, fastmath: bool = False):
        """Initialize a phase with given state, control and static variables.

        If names are given, they are used as the names of the variables.
        Otherwise, the names are generated automatically as x_0 through x_{n - 1}.

        static variables should be identical with which defined in the system.

        identifier should be unique for each phase in a given system to avoid possible name conflict.

        It is recommended to use method of System object to create phase, instead of manually.

        If simplify is True, every symbolic expression will be simplified (by sympy.simplify)
        before being compiled. This may slow down the speed of compilation.

        If parallel is True, parallel flag will be passed to numba JIT compiler,
        which will generate parallel code for multicore CPUs.
        This may slow down the speed of compilation.

        If fastmath is True, fastmath flag will be passed to numba JIT compiler,
        see numba & LLVM documentations for details.

        Args:
            state (Union[int, List[str]]): Number of state variables or list of state variable names.
            control (Union[int, List[str]]): Number of control variables or list of control variable names.
            symbol_static_parameter (List[sp.Expr]): List of static parameters, should be identical to which in the system.
            identifier (int): Index of the phase.
            simplify (bool, optional): Whether to use sympy to simplify Exprs. Defaults to False.
            parallel (bool, optional): Whether to use numba parallel mode. Defaults to False.
            fastmath (bool, optional): Whether to use numba fastmath mode. Defaults to False.
        """
        self._identifier = identifier

        if isinstance(state, int):
            self._num_state = state
            self._name_state = [f'x_{i}^{{({identifier})}}' for i in range(state)]
        elif isinstance(state, list):
            self._name_state = [s + f'^{{({identifier})}}' for s in state]
            self._num_state = len(state)
        else:
            raise ValueError('state must be int or list of str')

        if isinstance(control, int):
            self._num_control = control
            self._name_control = [f'u_{i}^{{({identifier})}}' for i in range(control)]
        elif isinstance(control, list):
            self._name_control = [c + f'^{{({identifier})}}' for c in control]
            self._num_control = len(control)
        else:
            raise ValueError('control must be int or list of str')

        self._num_static_parameter = len(symbol_static_parameter)
        self._name_static_parameter = [p.name for p in symbol_static_parameter]
        self._symbol_static_parameter = symbol_static_parameter

        self._symbol_state = [sp.Symbol(name) for name in self._name_state]
        self._symbol_control = [sp.Symbol(name) for name in self._name_control]
        self._symbol_time = sp.Symbol(f't^{{({identifier})}}')
        self._symbols = self._symbol_state + self._symbol_control + symbol_static_parameter + [self._symbol_time]

        self._simplify = simplify
        self._parallel = parallel
        self._fastmath = fastmath

        self._dynamics_set = False  # user must set dynamics
        self._boundary_condition_set = False  # user must set boundary conditions
        self._discretization_set = False  # user must set a discretization scheme
        self._integral_set = False
        self._phase_constraint_set = False
        self.set_integral([])  # no integral by default
        self.set_phase_constraint([], [], [])  # no phase constraint by default

    def set_dynamics(self, dynamics: List[Union[float, sp.Expr]]):
        """Set the dynamics of the phase.

        Args:
            dynamics (List[sp.Expr]): List of derivatives of states composed with x, u, s, and t.
        """
        if not len(dynamics) == self._num_state:
            raise ValueError('number of dynamics must be same as number of state variables')

        self._expr_dynamics = [sp.sympify(d) for d in dynamics]
        self._func_dynamics = [FastFunc(d, self._symbols, self._simplify, self._parallel, self._fastmath)
                               for d in self._expr_dynamics]
        self._dynamics_set = True  # must be the last line of this method, to avoid inconsistency if runtime error occurs
        return self

    def set_integral(self, integral: List[Union[float, sp.Expr]]):
        """Set the integrals of the phase.
        Symbols I_0, ..., I_{n - 1} will be automatically generated to be used to represent
        corresponding integrals in system.

        Args:
            integral (List[sp.Expr]): List of integrals to be concerned composed with x, u, s, and t.
        """
        self._expr_integral = [sp.sympify(i) for i in integral]
        self._func_integral = [FastFunc(i, self._symbols, self._simplify, self._parallel, self._fastmath)
                               for i in self._expr_integral]
        self._num_integral = len(integral)
        self._symbol_integral = [sp.Symbol(f'I_{i}^{{({self._identifier})}}') for i in range(self._num_integral)]

        self._integral_set = True
        return self

    def set_phase_constraint(self, phase_constraint: List[Union[float, sp.Expr]],
                             lower_bound: Iterable[Union[float]], upper_bound: Iterable[Union[float]]):
        """Set the phase constraint of the system, which is enforced in the entire phase.
        for equality constraint, set lower bound and upper bound to the same value.

        Args:
            phase_constraint (List[sp.Expr]): List of phase constraints composed with x, u, s, t
        """
        phase_constraint = list(phase_constraint)
        lower_bound = list(lower_bound)
        upper_bound = list(upper_bound)
        if not len(phase_constraint) == len(lower_bound) == len(upper_bound):
            raise ValueError('phase_constraint, lower_bound and upper_bound must have the same length')

        self._variable_bounds_phase = []
        self._static_parameter_bounds_phase = []
        self._time_bounds_phase = []

        self._expr_phase_constraint = []
        lower_bound_phase_constraint = []
        upper_bound_phase_constraint = []
        for c, lb, ub in zip(phase_constraint, lower_bound, upper_bound):
            if c.is_symbol:
                i = self._symbols.index(c)
                if i < self.n:
                    self._variable_bounds_phase.append((i, lb, ub))
                elif i < self.n + self.n_s:
                    self._static_parameter_bounds_phase.append((i - self.n, lb, ub))
                else:
                    self._time_bounds_phase.append((lb, ub))
            else:
                self._expr_phase_constraint.append(sp.sympify(c))
                lower_bound_phase_constraint.append(lb)
                upper_bound_phase_constraint.append(ub)
        self._func_phase_constraint = [FastFunc(c, self._symbols, self._simplify, self._parallel, self._fastmath)
                                       for c in self._expr_phase_constraint]

        self._num_phase_constraint = len(self._expr_phase_constraint)
        self._lower_bound_phase_constraint = np.array(lower_bound_phase_constraint, dtype=np.float64)
        self._upper_bound_phase_constraint = np.array(upper_bound_phase_constraint, dtype=np.float64)

        self._phase_constraint_set = True  # set flag to true **before** calling _update*

        if self._boundary_condition_set and self._discretization_set:
            self._update_bound_variable()

        return self

    def set_boundary_condition(self, initial_value: List[Union[None, float, sp.Expr]],
                               terminal_value: List[Union[None, float, sp.Expr]],
                               initial_time: Union[None, float, sp.Expr],
                               terminal_time: Union[None, float, sp.Expr]):
        """Set the boundary condition & initial/terminal time of the phase. None for free, otherwise fixed.
         Set it to sympy.Expr of static parameters to enforce relations between difference phases.

        Args:
            initial_value (List[Union[None, float]]): List of initial values of states.
            terminal_value (List[Union[None, float]]): List of terminal values of states.
            initial_time (Union[None, float]): Initial time.
            terminal_time (Union[None, float]): Terminal time.
        """
        if not len(initial_value) == len(terminal_value) == self._num_state:
            raise ValueError('initial_value, terminal_value must have the same length as number of state variables')
        for i in range(self.n_x):
            if isinstance(initial_value[i], int):
                initial_value[i] = float(initial_value[i])
            if isinstance(terminal_value[i], int):
                terminal_value[i] = float(terminal_value[i])
        if isinstance(initial_time, int):
            initial_time = float(initial_time)
        if isinstance(terminal_time, int):
            terminal_time = float(terminal_time)

        self._initial_value = initial_value
        self._terminal_value = terminal_value
        self._initial_time = initial_time
        self._terminal_time = terminal_time

        self._expr_boundary_condition = []

        # x0, xf, t0, tf, respectively. records to be used in set_discretization
        self._mapping_boundary_condition = [[], [], False, False]
        for i, v in enumerate(initial_value):
            if isinstance(v, sp.Expr):
                self._mapping_boundary_condition[0].append(i)
                self._expr_boundary_condition.append(v)
        for i, v in enumerate(terminal_value):
            if isinstance(v, sp.Expr):
                self._mapping_boundary_condition[1].append(i)
                self._expr_boundary_condition.append(v)
        if isinstance(initial_time, sp.Expr):
            self._mapping_boundary_condition[2] = True
            self._expr_boundary_condition.append(initial_time)
        if isinstance(terminal_time, sp.Expr):
            self._mapping_boundary_condition[3] = True
            self._expr_boundary_condition.append(terminal_time)

        self._func_boundary_condition = [FastFunc(bc, self._symbol_static_parameter, self._simplify, self._parallel,
                                                  self._fastmath) for bc in self._expr_boundary_condition]

        self._boundary_condition_set = True

        if self._discretization_set:
            self._update_boundary_condition_index()

        if self._discretization_set and self._phase_constraint_set:
            self._update_bound_variable()
        return self

    def set_discretization(self, mesh: Union[int, Iterable[float]], num_point: Union[int, Iterable[int]],
                           state_continuity: Union[bool, Iterable[bool]],
                           control_continuity: Union[bool, Iterable[bool]]):
        """Set the discretization scheme of the system.

        Mesh will be rescaled when needed. If it is set with an integer, uniform mesh is used.
        If it is set with an array, the array will be used as mesh directly.

        num_point decides the number of interpolation / integration points in each mesh interval.
        If it is set with an integer, the same number of points will be used in each mesh interval.

        state_continuous and control_continuous decides whether the state and control are continuous in the phase.
        E.g., in Bang-Bang control problem, set the control_continuous to False for corresponding control variables.
        state & control continuous flags must be correctly set to ensure variables have proper discretization scheme.
        If it is set with a boolean, the same continuity will be used in all states or controls.

        Args:
            mesh (Union[int, NDArray[np.float64]]): Number of mesh or mesh points.
            num_point (Union[int, NDArray[np.int32]]): Number of points in each mesh.
            state_continuity (Union[bool, List[bool]]): List of bools indicating whether the state is continuous or not.
            control_continuity (Union[bool, List[bool]]): List of bools indicating whether the control is continuous or not.
        """
        if isinstance(mesh, int):  # uniform mesh
            self._mesh = np.linspace(0, 1, mesh + 1, endpoint=True)
        else:  # scale to [0, 1]
            mesh = np.array(list(mesh), dtype=np.float64)
            self._mesh = (mesh - mesh[0]) / (mesh[-1] - mesh[0])
        self._num_interval = len(self._mesh) - 1

        if isinstance(num_point, int):
            self._num_point = np.full(self._num_interval, num_point, dtype=np.int32)
        else:
            num_point = np.array(list(num_point), dtype=np.int32)
            self._num_point = num_point

        if len(self._num_point) != self._num_interval:
            raise ValueError('num_point must have the same length as mesh intervals (= len(mesh) - 1)')

        if isinstance(state_continuity, bool):
            self._continuity_state = np.full(self._num_state, state_continuity, dtype=np.bool)
        else:
            self._continuity_state = np.array(list(state_continuity), dtype=np.bool)

        if isinstance(control_continuity, bool):
            self._continuity_control = np.full(self._num_control, control_continuity, dtype=np.bool)
        else:
            self._continuity_control = np.array(list(control_continuity), dtype=np.bool)

        self._continuity = np.concatenate((self._continuity_state, self._continuity_control))

        self.l_c, self.r_c = lr_c(self._num_point)
        self.l_nc, self.r_nc = lr_nc(self._num_point)
        self.L_c = self.r_c[-1]
        self.L_nc = self.r_nc[-1]

        self.t_c, self.w_c = xw_c(self._mesh, self._num_point)
        self.t_nc, self.w_nc = xw_nc(self._mesh, self._num_point)

        self.f_c2nc = c2nc(self._num_point)
        self.f_nc2c = nc2c(self._num_point)
        self.D_c = D_c(self._mesh, self._num_point)
        self.D_nc = D_nc(self._mesh, self._num_point)

        self.l_v, self.r_v = lr_v(self._num_point, self._continuity)
        self.l_m, self.r_m = lr_m(self._num_point, self._continuity)
        self.L_v = self.r_v[-1]
        self.L_m = self.r_m[-1]

        self.f_v2m = v2m(self._num_point, self._continuity)
        self.f_m2v = m2v(self._num_point, self._continuity)
        self.D_x = D_x(self._mesh, self._num_point, self._continuity_state)

        self.D_coo = self.D_x.tocoo()  # cache to speed up jacobian & hessian calculation
        self.D_coo.eliminate_zeros()

        self._discretization_set = True

        if self._boundary_condition_set:
            self._update_boundary_condition_index()
        if self._boundary_condition_set and self._phase_constraint_set:
            self._update_bound_variable()
        return self

    def _update_bound_variable(self):
        """Update lower and upper bound of variables.

        Should be called after discretization scheme, phase constraints, and boundary condition is set.
        """
        self._lower_bound_variable = np.full(self.L, -np.inf, dtype=np.float64)
        self._upper_bound_variable = np.full(self.L, np.inf, dtype=np.float64)
        for i, lb, ub in self._variable_bounds_phase:
            self._lower_bound_variable[self.l_v[i]:self.r_v[i]] \
                = np.maximum(self._lower_bound_variable[self.l_v[i]:self.r_v[i]], lb)
            self._upper_bound_variable[self.l_v[i]:self.r_v[i]] \
                = np.minimum(self._upper_bound_variable[self.l_v[i]:self.r_v[i]], ub)
        for lb, ub in self._time_bounds_phase:
            self._lower_bound_variable[-2] = np.maximum(self._lower_bound_variable[-2], lb)
            self._lower_bound_variable[-1] = np.maximum(self._lower_bound_variable[-1], lb)
            self._upper_bound_variable[-2] = np.minimum(self._upper_bound_variable[-2], ub)
            self._upper_bound_variable[-1] = np.minimum(self._upper_bound_variable[-1], ub)
        for i in range(self.n_x):
            if isinstance(self.bc_0[i], float):
                self._lower_bound_variable[self.l_v[i]] \
                    = np.maximum(self._lower_bound_variable[self.l_v[i]], self.bc_0[i])
                self._upper_bound_variable[self.l_v[i]] \
                    = np.minimum(self._upper_bound_variable[self.l_v[i]], self.bc_0[i])
            if isinstance(self.bc_f[i], float):
                self._lower_bound_variable[self.r_v[i] - 1] \
                    = np.maximum(self._lower_bound_variable[self.r_v[i] - 1], self.bc_f[i])
                self._upper_bound_variable[self.r_v[i] - 1] \
                    = np.minimum(self._upper_bound_variable[self.r_v[i] - 1], self.bc_f[i])
        if isinstance(self.t_0, float):
            self._lower_bound_variable[-2] = np.maximum(self._lower_bound_variable[-2], self.t_0)
            self._upper_bound_variable[-2] = np.minimum(self._upper_bound_variable[-2], self.t_0)
        if isinstance(self.t_f, float):
            self._lower_bound_variable[-1] = np.maximum(self._lower_bound_variable[-1], self.t_f)
            self._upper_bound_variable[-1] = np.minimum(self._upper_bound_variable[-1], self.t_f)

    def _update_boundary_condition_index(self):
        """Update the index of boundary conditions in the discretized system.

        Should be called after both discretization scheme and boundary condition is set.
        """
        index_boundary_condition = []
        for i in self._mapping_boundary_condition[0]:
            index_boundary_condition.append(self.l_v[i])
        for i in self._mapping_boundary_condition[1]:
            index_boundary_condition.append(self.r_v[i] - 1)
        if self._mapping_boundary_condition[2]:
            index_boundary_condition.append(self.L_v)
        if self._mapping_boundary_condition[3]:
            index_boundary_condition.append(self.L_v + 1)
        self._index_boundary_condition = np.array(index_boundary_condition, dtype=np.int32)
        self._num_boundary_condition = len(self._index_boundary_condition)

    @property
    def n_x(self) -> int:
        """Number of states."""
        return self._num_state

    @property
    def x(self) -> List[sp.Symbol]:
        """Sympy symbols of states."""
        return self._symbol_state

    @property
    def c_x(self) -> NDArray[np.bool]:
        """Continuity of states."""
        return self._continuity_state

    @property
    def n_u(self) -> int:
        """Number of controls."""
        return self._num_control

    @property
    def u(self) -> List[sp.Symbol]:
        """Sympy symbols of controls."""
        return self._symbol_control

    @property
    def c_u(self) -> NDArray[np.bool]:
        """Continuity of controls."""
        return self._continuity_control

    @property
    def n(self) -> int:
        """Number of states and controls."""
        return self._num_state + self._num_control

    @property
    def c(self) -> NDArray[np.bool]:
        """Continuity of states and controls."""
        return self._continuity

    @property
    def n_s(self) -> int:
        """Number of static parameters."""
        return self._num_static_parameter

    @property
    def s(self) -> List[sp.Symbol]:
        """Sympy symbols of static parameters."""
        return self._symbol_static_parameter

    @property
    def t(self):
        """Sympy symbols of time."""
        return self._symbol_time

    @property
    def F_d(self) -> List[FastFunc]:
        """FastFuncs of dynamics."""
        return self._func_dynamics

    @property
    def n_d(self) -> int:
        """Number of dynamics."""
        return self._num_state

    @property
    def F_I(self) -> List[FastFunc]:
        """FastFuncs of integrals."""
        return self._func_integral

    @property
    def n_I(self) -> int:
        """Number of integrals."""
        return self._num_integral

    @property
    def I(self) -> List[sp.Symbol]:
        """Sympy symbols of integrals."""
        return self._symbol_integral

    @property
    def F_c(self) -> List[FastFunc]:
        """FastFuncs of constraints."""
        return self._func_phase_constraint

    @property
    def n_c(self) -> int:
        """Number of constraints."""
        return self._num_phase_constraint

    @property
    def v_lb(self) -> NDArray[np.float64]:
        """Lower bounds of variables."""
        return self._lower_bound_variable

    @property
    def v_ub(self) -> NDArray[np.float64]:
        """Upper bounds of variables."""
        return self._upper_bound_variable

    @property
    def c_lb(self) -> NDArray[np.float64]:
        """Lower bounds of constraints."""
        return self._lower_bound_phase_constraint

    @property
    def c_ub(self) -> NDArray[np.float64]:
        """Upper bounds of constraints."""
        return self._upper_bound_phase_constraint

    @property
    def s_b(self) -> List[Tuple[int, float, float]]:
        """Bounds of static parameters."""
        return self._static_parameter_bounds_phase

    @property
    def F_bc(self) -> List[FastFunc]:
        """FastFuncs of boundary conditions which is sympy.expr."""
        return self._func_boundary_condition

    @property
    def n_bc(self) -> int:
        """Number of boundary conditions which is sympy.expr."""
        return self._num_boundary_condition

    @property
    def i_bc(self) -> NDArray[np.float64]:
        """Index of boundary conditions which is sympy.expr."""
        return self._index_boundary_condition

    @property
    def bc_0(self) -> List[Union[None, int, float, sp.Expr]]:
        """Initial boundary conditions."""
        return self._initial_value

    @property
    def bc_f(self) -> List[Union[None, int, float, sp.Expr]]:
        """Terminal boundary conditions."""
        return self._terminal_value

    @property
    def t_0(self) -> Union[None, int, float, sp.Expr]:
        """Initial time."""
        return self._initial_time

    @property
    def t_f(self) -> Union[None, int, float, sp.Expr]:
        """Terminal time."""
        return self._terminal_time

    @property
    def N(self) -> int:
        """Number of mesh intervals."""
        return self._num_interval

    @property
    def L(self) -> int:
        """Length of optimization variables, including states, controls, and time."""
        return self.r_v[-1] + 2

    @property
    def L_x(self) -> int:
        """Length of optimization variables of states."""
        return self.r_v[self.n_x - 1]

    @property
    def L_xu(self) -> int:
        """Length of optimization variables of states and controls."""
        return self.r_v[-1]

    @property
    def l0_m(self) -> int:
        """Length of single variable in middle stage"""
        return self.L_nc

    @property
    def t_m(self) -> NDArray[np.float64]:
        """t in middle stage"""
        return self.t_nc

    @property
    def w(self) -> NDArray[np.float64]:
        """Integration weights is integration weights for non-continuous variables."""
        return self.w_nc

    @property
    def ok(self):
        """Whether the phase is fully configured."""
        return self._dynamics_set and self._boundary_condition_set and self._discretization_set



from typing import List, Union, Iterable

import cyipopt
import numpy as np

from .system import System
from .variable import Variable


def ipopt_solve(system: System, guess: Union[Variable, List[Union[Variable, Iterable[float]]]],
                optimizer_options: dict = None):
    """Solve the system using IPOPT. If the system has only one phase and no static variables,
     guess can be a single variable. Otherwise, guess should be a list of variables, one for each phase,
     appended with a list of guess values for static variables. Optimizer options should be a dictionary
     of options to pass to Ipopt. See https://coin-or.github.io/Ipopt/OPTIONS.html. Options will be passed
     verbatim to IPOPT.

     Returns the value returned by IPOPT parsed as the same format as guess (a single Variable or a list of
     Variables and static values), and a dictionary of information about the solution returned by IPOPT.
    """
    if not system.ok:
        raise ValueError("system is not fully configured")
    if optimizer_options is None:
        optimizer_options = {}

    guess_is_variable = isinstance(guess, Variable)
    if guess_is_variable:
        guess = [guess]

    x_0 = np.zeros(system.L)
    for i in range(system.n_p):
        x_0[system.l_p[i]:system.r_p[i]] = guess[i].data
    if system.n_s > 0:
        x_0[system.l_s:system.r_s] = np.array(list(guess[-1]), dtype=np.float64)

    solver = cyipopt.Problem(n=int(system.L), m=(len(system.c_lb)), problem_obj=system, lb=system.v_lb, ub=system.v_ub,
                             cl=system.c_lb, cu=system.c_ub)

    for k, v in optimizer_options.items():
        solver.add_option(k, v)

    x, info = solver.solve(x_0)
    result = []
    for i in range(system.n_p):
        result.append(Variable(system.p[i], x[system.l_p[i]:system.r_p[i]]))
    if system.n_s > 0:
        result.append(x[system.l_s:system.r_s])

    if guess_is_variable:
        return result[0], info
    return result, info



import numpy as np
from numpy.typing import NDArray

from .phase import Phase


class BatchIndexArray:
    def __init__(self, data: NDArray[np.float64], l_ind: NDArray[np.int32], r_ind: NDArray[np.int32]) -> None:
        super().__init__()
        if not len(l_ind) == len(r_ind):
            raise ValueError("l_ind and r_ind must have the same length")
        self._data = data
        self._l_ind = l_ind
        self._r_ind = r_ind
        self._n = len(l_ind)

    def __getitem__(self, i: int) -> NDArray[np.float64]:
        return self._data[self._l_ind[i]:self._r_ind[i]]

    def __setitem__(self, i: int, value: NDArray[np.float64]) -> None:
        self._data[self._l_ind[i]:self._r_ind[i]] = value

    def __len__(self) -> int:
        return self._n


class Variable:
    def __init__(self, phase: Phase, data: NDArray[np.float64]) -> None:
        self._data = data
        self._l_v = phase.l_v
        self._r_v = phase.r_v
        self._t_c = phase.t_c
        self._t_nc = phase.t_nc
        self._n_x = phase.n_x
        self._n = phase.n
        self._array_state = BatchIndexArray(data, self._l_v[:self._n_x], self._r_v[:self._n_x])
        self._array_control = BatchIndexArray(data, self._l_v[self._n_x:], self._r_v[self._n_x:])

    @property
    def x(self) -> BatchIndexArray:
        return self._array_state

    @property
    def u(self) -> BatchIndexArray:
        return self._array_control

    @property
    def t_0(self) -> np.float64:
        return self._data[-2]

    @t_0.setter
    def t_0(self, value: np.float64) -> None:
        self._data[-2] = value

    @property
    def t_f(self) -> np.float64:
        return self._data[-1]

    @t_f.setter
    def t_f(self, value: np.float64) -> None:
        self._data[-1] = value

    @property
    def t_c(self) -> NDArray[np.float64]:
        return self._t_c * (self.t_f - self.t_0) + self.t_0

    @property
    def t_nc(self) -> NDArray[np.float64]:
        return self._t_nc * (self.t_f - self.t_0) + self.t_0

    @property
    def data(self) -> NDArray[np.float64]:
        return self._data


def constant_guess(phase: Phase, value: np.float = 1.) -> Variable:
    if not phase.ok:
        raise ValueError("phase is not fully configured")
    v = Variable(phase, np.full(phase.L, value))
    for i in range(phase.n_x):
        if isinstance(phase.bc_0[i], float):
            v.x[i][0] = phase.bc_0[i]
        elif isinstance(phase.bc_f[i], float):
            v.x[i][-1] = phase.bc_f[i]
    if isinstance(phase.t_0, float):
        v.t_0 = phase.t_0
    if isinstance(phase.t_f, float):
        v.t_f = phase.t_f
    return v


def linear_guess(phase: Phase, default: np.float = 1.) -> Variable:
    if not phase.ok:
        raise ValueError("phase is not fully configured")
    v = Variable(phase, np.full(phase.L, default))
    if isinstance(phase.t_0, float):
        v.t_0 = phase.t_0
    if isinstance(phase.t_f, float):
        v.t_f = phase.t_f

    for i in range(phase.n_x):
        if isinstance(phase.bc_0[i], float) and isinstance(phase.bc_f[i], float):
            if phase.c_x[i]:
                v.x[i] = v.t_c * (phase.bc_f[i] - phase.bc_0[i]) + phase.bc_0[i]
            else:
                v.x[i] = v.t_nc * (phase.bc_f[i] - phase.bc_0[i]) + phase.bc_0[i]
        elif isinstance(phase.bc_0[i], float):
            if phase.c_x[i]:
                v.x[i] = v.t_c * (default - phase.bc_0[i]) + phase.bc_0[i]
            else:
                v.x[i] = v.t_nc * (default - phase.bc_0[i]) + phase.bc_0[i]
        elif isinstance(phase.bc_f[i], float):
            if phase.c_x[i]:
                v.x[i] = v.t_c * (phase.bc_f[i] - default) + default
            else:
                v.x[i] = v.t_nc * (phase.bc_f[i] - default) + default
    return v

